@@ -1,12 +1,38 @@
+/*
+ * Copyright (c) 1984 through 2007, William LeFebvre
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *     * Redistributions of source code must retain the above copyright
+ * notice, this list of conditions and the following disclaimer.
+ * 
+ *     * Redistributions in binary form must reproduce the above
+ * copyright notice, this list of conditions and the following disclaimer
+ * in the documentation and/or other materials provided with the
+ * distribution.
+ * 
+ *     * Neither the name of William LeFebvre nor the names of other
+ * contributors may be used to endorse or promote products derived from
+ * this software without specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  *  Top users/processes display for Unix
  *  Version 3
- *
- *  This program may be freely redistributed,
- *  but this entire comment MUST remain intact.
- *
- *  Copyright (c) 1984, 1989, William LeFebvre, Rice University
- *  Copyright (c) 1989, 1990, 1992, William LeFebvre, Northwestern University
  */
 
 /*
@@ -21,14 +47,17 @@
  *        In this way, those routines can be safely used on terminals that
  *        have minimal (or nonexistant) terminal capabilities.
  *
- *        The routines are called in this order:  *_loadave, i_timeofday,
- *        *_procstates, *_cpustates, *_memory, *_message, *_header,
- *        *_process, u_endscreen.
+ *        The routines should be called in this order:  *_loadave, *_uptime,
+ *        i_timeofday, *_procstates, *_cpustates, *_memory, *_swap,
+ *        *_message, *_header, *_process, *_endscreen.
  */
 
 #include "os.h"
 #include <ctype.h>
 #include <stdarg.h>
+#include <sys/types.h>
+#include <sys/uio.h>
+#include <unistd.h>
 
 #include "top.h"
 #include "machine.h"
@@ -44,6 +73,8 @@
 
 #define CURSOR_COST 8
 
+#define MESSAGE_DISPLAY_TIME 5
+
 /* imported from screen.c */
 extern int overstrike;
 
@@ -64,14 +95,14 @@ static int x_uptime = X_UPTIME;
 static int y_uptime = Y_UPTIME;
 static int x_procstate = X_PROCSTATE;
 static int y_procstate = Y_PROCSTATE;
-static int x_brkdn = X_BRKDN;
-static int y_brkdn = Y_BRKDN;
 static int x_cpustates = X_CPUSTATES;
 static int y_cpustates = Y_CPUSTATES;
+static int x_kernel = X_KERNEL;
+static int y_kernel = Y_KERNEL;
 static int x_mem = X_MEM;
 static int y_mem = Y_MEM;
-static int x_swap = -1;
-static int y_swap = -1;
+static int x_swap = X_SWAP;
+static int y_swap = Y_SWAP;
 static int y_message = Y_MESSAGE;
 static int x_header = X_HEADER;
 static int y_header = Y_HEADER;
@@ -102,11 +133,13 @@ static char **procstate_names;
 static char **cpustate_names;
 static char **memory_names;
 static char **swap_names;
+static char **kernel_names;
 
 static int num_procstates;
 static int num_cpustates;
 static int num_memory;
 static int num_swap;
+static int num_kernel;
 
 static int *lprocstates;
 static int *lcpustates;
@@ -114,7 +147,24 @@ static int *lcpustates;
 static int *cpustate_columns;
 static int cpustate_total_length;
 
-static enum { OFF, ON, ERASE } header_status = ON;
+static int header_status = Yes;
+
+/* pending messages are stored in a circular buffer, where message_first
+   is the next one to display, and message_last is the last one
+   in the buffer.  Counters wrap around at MAX_MESSAGES.  The buffer is
+   empty when message_first == message_last and full when 
+   message_last + 1 == message_first.  The pointer message_current holds
+   the message currently being displayed, or "" if there is none.
+*/
+#define MAX_MESSAGES 16
+static char *message_buf[MAX_MESSAGES];
+static int message_first = 0;
+static int message_last = 0;
+static struct timeval message_time = {0, 0};
+static char *message_current = NULL;
+static int message_length = 0;
+static int message_hold = 1;
+static int message_barrier = No;
 
 #ifdef ENABLE_COLOR
 static int load_cidx[3];
@@ -122,8 +172,12 @@ static int header_cidx;
 static int *cpustate_cidx;
 static int *memory_cidx;
 static int *swap_cidx;
+static int *kernel_cidx;
+#else
+#define memory_cidx NULL
+#define swap_cidx NULL
+#define kernel_cidx NULL
 #endif
-static int header_color = 0;
 
 
 /* internal support routines */
@@ -157,7 +211,7 @@ display_clear()
 
 {
     dprintf("display_clear\n");
-    clear();
+    screen_clear();
     memzero(screenbuf, bufsize);
     memzero(colorbuf, bufsize);
     curr_x = curr_y = 0;
@@ -196,12 +250,14 @@ display_move(int x, int y)
 	/* one newline for every line */
 	while (cnt > 0 && curr_y < y)
 	{
+#ifdef ENABLE_COLOR
 	    if (color != 0)
 	    {
-		p = strecpy(p, color_set(0));
+		p = strcpyend(p, color_setstr(0));
 		color = 0;
 		cnt -= 5;
 	    }
+#endif
 	    *p++ = '\n';
 	    curr_y++;
 	    curr_x = 0;
@@ -213,12 +269,14 @@ display_move(int x, int y)
 	colorp = &colorbuf[lineindex(curr_y) + curr_x];
 	while (cnt > 0 && curr_x < x)
 	{
+#ifdef ENABLE_COLOR
 	    if (color != *colorp)
 	    {
 		color = *colorp;
-		p = strecpy(p, color_set(color));
+		p = strcpyend(p, color_setstr(color));
 		cnt -= 5;
 	    }
+#endif
 	    if ((*p = *bufp) == '\0')
 	    {
 		/* somwhere on screen we haven't been before */
@@ -242,7 +300,7 @@ display_move(int x, int y)
     }
     else
     {
-	Move_to(x, y);
+	screen_move(x, y);
     }
 
     /* update our position */
@@ -336,11 +394,13 @@ display_write(int x, int y, int newcolor, int eol, char *new)
 	    }
 
 	    /* write character */
+#ifdef ENABLE_COLOR
 	    if (curr_color != newcolor)
 	    {
-		fputs(color_set(newcolor), stdout);
+		fputs(color_setstr(newcolor), stdout);
 		curr_color = newcolor;
 	    }
+#endif
 	    putchar(ch);
 	    *bufp = ch;
 	    *colorp = curr_color;
@@ -359,22 +419,24 @@ display_write(int x, int y, int newcolor, int eol, char *new)
     {
 	dprintf("display_write: clear-eol (bufp = \"%s\")\n", bufp);
 	/* make sure we are color 0 */
+#ifdef ENABLE_COLOR
 	if (curr_color != 0)
 	{
-	    fputs(color_set(0), stdout);
+	    fputs(color_setstr(0), stdout);
 	    curr_color = 0;
 	}
+#endif
 
 	/* make sure we are at the end */
 	if (x != curr_x || y != curr_y)
 	{
-	    Move_to(x, y);
+	    screen_move(x, y);
 	    curr_x = x;
 	    curr_y = y;
 	}
 
 	/* clear to end */
-	clear_eol(strlen(bufp));
+	screen_cleareol(strlen(bufp));
 
 	/* clear out whats left of this line's buffer */
 	diff = display_width - x;
@@ -431,51 +493,53 @@ display_cte()
     {
 	dprintf("display_cte: clearing\n");
 
-	/* different method when there's no clear_to_end */
-	if (clear_to_end)
+	/* we will need this later */
+	len = lineindex(virt_y) + virt_x;
+
+	/* move to x and y, then clear to end */
+	display_move(virt_x, virt_y);
+	if (!screen_cte())
 	{
-	    display_move(virt_x, virt_y);
-	    putcap(clear_to_end);
-	}
-	else
-	{
-	    if (++virt_y < screen_length)
+	    /* screen has no clear to end, so do it by hand */
+	    p = &screenbuf[len];
+	    len = strlen(p);
+	    if (len > 0)
+	    {
+		screen_cleareol(len);
+	    }
+	    while (++virt_y < screen_length)
 	    {
 		display_move(0, virt_y);
-		virt_x = 0;
-		while (virt_y < screen_length)
+		p = &screenbuf[lineindex(virt_y)];
+		len = strlen(p);
+		if (len > 0)
 		{
-		    p = &screenbuf[lineindex(virt_y)];
-		    len = strlen(p);
-		    if (len > 0)
-		    {
-			clear_eol(len);
-		    }
-		    virt_y++;
+		    screen_cleareol(len);
 		}
 	    }
 	}
 
 	/* clear the screenbuf */
-	len = lineindex(virt_y) + virt_x;
-	memzero(&screenbuf[len], bufsize -len);
-	memzero(&colorbuf[len], bufsize -len);
+	memzero(&screenbuf[len], bufsize - len);
+	memzero(&colorbuf[len], bufsize - len);
     }
 }
 
 static void
-summary_format(int x, int y, int *numbers, char **names)
+summary_format(int x, int y, int *numbers, char **names, int *cidx)
 
 {
     register int num;
     register char *thisname;
     register char *lastname = NULL;
+    register int color;
 
     /* format each number followed by its string */
     while ((thisname = *names++) != NULL)
     {
 	/* get the number to format */
 	num = *numbers++;
+	color = 0;
 
 	/* display only non-zero numbers */
 	if (num != 0)
@@ -486,10 +550,18 @@ summary_format(int x, int y, int *numbers, char **names)
 		display_write(-1, -1, 0, 0, lastname);
 	    }
 
+#ifdef ENABLE_COLOR
+	    if (cidx != NULL)
+	    {
+		/* choose a color */
+		color = color_test(*cidx++, num);
+	    }
+#endif
+
 	    /* write this number if positive */
 	    if (num > 0)
 	    {
-		display_write(x, y, 0, 0, itoa(num));
+		display_write(x, y, color, 0, itoa(num));
 	    }
 
 	    /* defer writing this name */
@@ -502,14 +574,17 @@ summary_format(int x, int y, int *numbers, char **names)
 
     /* if the last string has a separator on the end, it has to be
        written with care */
-    if ((num = strlen(lastname)) > 1 &&
-	lastname[num-2] == ',' && lastname[num-1] == ' ')
+    if (lastname != NULL)
     {
-	display_fmt(-1, -1, 0, 1, "%.*s", num-2, lastname);
-    }
-    else
-    {
-	display_write(-1, -1, 0, 1, lastname);
+	if ((num = strlen(lastname)) > 1 &&
+	    lastname[num-2] == ',' && lastname[num-1] == ' ')
+	{
+	    display_fmt(-1, -1, 0, 1, "%.*s", num-2, lastname);
+	}
+	else
+	{
+	    display_write(-1, -1, 0, 1, lastname);
+	}
     }
 }
 
@@ -588,12 +663,12 @@ int
 display_resize()
 
 {
-    register int lines;
+    register int top_lines;
     register int newsize;
 
     /* calculate the current dimensions */
     /* if operating in "dumb" mode, we only need one line */
-    lines = smart_terminal ? screen_length : 1;
+    top_lines = smart_terminal ? screen_length : 1;
 
     /* we don't want more than MAX_COLS columns, since the machine-dependent
        modules make static allocations based on MAX_COLS and we don't want
@@ -605,7 +680,7 @@ display_resize()
     }
 
     /* see how much space we need */
-    newsize = lines * (MAX_COLS + 1);
+    newsize = top_lines * (MAX_COLS + 1);
 
     /* reallocate only if we need more than we already have */
     if (newsize > bufsize)
@@ -638,11 +713,18 @@ display_resize()
     }
 
     /* adjust total lines on screen to lines available for procs */
-    lines -= y_procs;
+    top_lines -= y_procs;
 
     /* return number of lines available */
     /* for dumb terminals, pretend like we can show any amount */
-    return(smart_terminal ? lines : Largest);
+    return(smart_terminal ? top_lines : Largest);
+}
+
+int
+display_lines()
+
+{
+    return(smart_terminal ? screen_length : Largest);
 }
 
 /*
@@ -657,7 +739,7 @@ int
 display_init(struct statics *statics)
 
 {
-    register int lines;
+    register int top_lines;
     register char **pp;
     register char *p;
     register int *ip;
@@ -665,6 +747,20 @@ display_init(struct statics *statics)
 
     /* certain things may influence the screen layout,
        so look at those first */
+
+    /* a kernel line shifts parts of the display down */
+    kernel_names = statics->kernel_names;
+    if ((num_kernel = string_count(kernel_names)) > 0)
+    {
+	/* adjust screen placements */
+	y_mem++;
+	y_swap++;
+	y_message++;
+	y_header++;
+	y_idlecursor++;
+	y_procs++;
+    }
+
     /* a swap line shifts parts of the display down one */
     swap_names = statics->swap_names;
     if ((num_swap = string_count(swap_names)) > 0)
@@ -674,15 +770,13 @@ display_init(struct statics *statics)
 	y_header++;
 	y_idlecursor++;
 	y_procs++;
-	x_swap = X_SWAP;
-	y_swap = Y_SWAP;
     }
     
     /* call resize to do the dirty work */
-    lines = display_resize();
+    top_lines = display_resize();
 
     /* only do the rest if we need to */
-    if (lines > -1)
+    if (top_lines > -1)
     {
 	/* save pointers and allocate space for names */
 	procstate_names = statics->procstate_names;
@@ -718,22 +812,34 @@ display_init(struct statics *statics)
 
     /* find header color */
     header_cidx = color_tag("header");
-    header_color = color_test(header_cidx, 0);
 
     /* color tags for cpu states */
     cpustate_cidx = (int *)malloc(num_cpustates * sizeof(int));
     i = 0;
-    p = strecpy(scratchbuf, "cpu.");
+    p = strcpyend(scratchbuf, "cpu.");
     while (i < num_cpustates)
     {
 	strcpy(p, cpustate_names[i]);
 	cpustate_cidx[i++] = color_tag(scratchbuf);
     }
 
+    /* color tags for kernel */
+    if (num_kernel > 0)
+    {
+	kernel_cidx = (int *)malloc(num_kernel * sizeof(int));
+	i = 0;
+	p = strcpyend(scratchbuf, "kernel.");
+	while (i < num_kernel)
+	{
+	    strcpy(p, homogenize(kernel_names[i]+1));
+	    kernel_cidx[i++] = color_tag(scratchbuf);
+	}
+    }
+
     /* color tags for memory */
     memory_cidx = (int *)malloc(num_memory * sizeof(int));
     i = 0;
-    p = strecpy(scratchbuf, "memory.");
+    p = strcpyend(scratchbuf, "memory.");
     while (i < num_memory)
     {
 	strcpy(p, homogenize(memory_names[i]+1));
@@ -741,18 +847,21 @@ display_init(struct statics *statics)
     }
 
     /* color tags for swap */
-    swap_cidx = (int *)malloc(num_swap * sizeof(int));
-    i = 0;
-    p = strecpy(scratchbuf, "swap.");
-    while (i < num_swap)
+    if (num_swap > 0)
     {
-	strcpy(p, homogenize(swap_names[i]+1));
-	swap_cidx[i++] = color_tag(scratchbuf);
+	swap_cidx = (int *)malloc(num_swap * sizeof(int));
+	i = 0;
+	p = strcpyend(scratchbuf, "swap.");
+	while (i < num_swap)
+	{
+	    strcpy(p, homogenize(swap_names[i]+1));
+	    swap_cidx[i++] = color_tag(scratchbuf);
+	}
     }
 #endif
 
     /* return number of lines available (or error) */
-    return(lines);
+    return(top_lines);
 }
 
 static void
@@ -775,9 +884,6 @@ i_loadave(int mpid, double *avenrun)
 {
     register int i;
 
-    /* i_loadave also clears the screen, since it is first */
-    display_clear();
-
     /* mpid == -1 implies this system doesn't have an _mpid */
     if (mpid != -1)
     {
@@ -896,11 +1002,23 @@ i_timeofday(time_t *tod)
      *  We want indices 11 thru 18 (length 8).
      */
 
-    display_fmt((smart_terminal ? screen_width : 79) - 8, 0, 0, 1,
-		"%-8.8s", &(ctime(tod)[11]));
+    int x;
+
+    /* where on the screen do we start? */
+    x = (smart_terminal ? screen_width : 79) - 8;
+
+    /* but don't bump in to uptime */
+    if (x < x_uptime + 19)
+    {
+	x = x_uptime + 19;
+    }
+
+    /* display it */
+    display_fmt(x, 0, 0, 1, "%-8.8s", &(ctime(tod)[11]));
 }
 
 static int ltotal = 0;
+static int lthreads = 0;
 
 /*
  *  *_procstates(total, brkdn, names) - print the process summary line
@@ -908,28 +1026,36 @@ static int ltotal = 0;
 
 
 void
-i_procstates(int total, int *brkdn)
+i_procstates(int total, int *brkdn, int threads)
 
 {
     /* write current number of processes and remember the value */
     display_fmt(0, y_procstate, 0, 0,
-		"%d processes: ", total);
+		"%d %s: ", total, threads ? "threads" : "processes");
     ltotal = total;
 
     /* remember where the summary starts */
     x_procstate = virt_x;
 
     /* format and print the process state summary */
-    summary_format(-1, -1, brkdn, procstate_names);
+    summary_format(-1, -1, brkdn, procstate_names, NULL);
 
     /* save the numbers for next time */
     memcpy(lprocstates, brkdn, num_procstates * sizeof(int));
+    lthreads = threads;
 }
 
 void
-u_procstates(int total, int *brkdn)
+u_procstates(int total, int *brkdn, int threads)
 
 {
+    /* if threads state has changed, do a full update */
+    if (lthreads != threads)
+    {
+	i_procstates(total, brkdn, threads);
+	return;
+    }
+
     /* update number of processes only if it has changed */
     if (ltotal != total)
     {
@@ -939,7 +1065,7 @@ u_procstates(int total, int *brkdn)
 	/* if number of digits differs, rewrite the label */
 	if (digits(total) != digits(ltotal))
 	{
-	    display_write(-1, -1, 0, 0, " processes: ");
+	    display_fmt(-1, -1, 0, 0, " %s: ", threads ? "threads" : "processes");
 	    x_procstate = virt_x;
 	}
 
@@ -951,7 +1077,7 @@ u_procstates(int total, int *brkdn)
     if (memcmp(lprocstates, brkdn, num_procstates * sizeof(int)) != 0)
     {
 	/* format and update the line */
-	summary_format(x_procstate, y_procstate, brkdn, procstate_names);
+	summary_format(x_procstate, y_procstate, brkdn, procstate_names, NULL);
 	memcpy(lprocstates, brkdn, num_procstates * sizeof(int));
     }
 }
@@ -1122,6 +1248,36 @@ z_cpustates()
     }
 }
 
+/*
+ *  *_kernel(stats) - print "Kernel: " followed by the kernel summary string
+ *
+ *  Assumptions:  cursor is on "lastline", the previous line
+ */
+
+void
+i_kernel(int *stats)
+
+{
+    if (num_kernel > 0)
+    {
+	display_write(0, y_kernel, 0, 0, "Kernel: ");
+
+	/* format and print the kernel summary */
+	summary_format(x_kernel, y_kernel, stats, kernel_names, kernel_cidx);
+    }
+}
+
+void
+u_kernel(int *stats)
+
+{
+    if (num_kernel > 0)
+    {
+	/* format the new line */
+	summary_format(x_kernel, y_kernel, stats, kernel_names, kernel_cidx);
+    }
+}
+
 /*
  *  *_memory(stats) - print "Memory: " followed by the memory summary string
  *
@@ -1190,41 +1346,89 @@ u_swap(long *stats)
 
 /*
  *  i_message is funny because it gets its message asynchronously (with
- *	respect to screen updates).
+ *	respect to screen updates).  Messages are taken out of the
+ *      circular message_buf and displayed one at a time.
  */
 
-static char next_msg[MAX_COLS + 8];
-static int msglen = 0;
-/* Invariant: msglen is always the length of the message currently displayed
-   on the screen (even when next_msg doesn't contain that message). */
-
 void
-i_message()
+i_message(struct timeval *now)
 
 {
+    struct timeval my_now;
+    int i = 0;
+
+    dprintf("i_message(%08x)\n", now);
+
+    /* if now is NULL we have to get it ourselves */
+    if (now == NULL)
+    {
+	time_get(&my_now);
+	now = &my_now;
+    }
+
+    /* now that we have been called, messages no longer need to be held */
+    message_hold = 0;
+
+    dprintf("i_message: now %d, message_time %d\n",
+	    now->tv_sec, message_time.tv_sec);
+
     if (smart_terminal)
     {
-	if (next_msg[0] != '\0')
+	/* is it time to change the message? */
+	if (timercmp(now, &message_time, > ))
 	{
-	    display_move(0, y_message);
-	    standout(next_msg);
-	    msglen = strlen(next_msg);
-	    next_msg[0] = '\0';
-	}
-	else if (msglen > 0)
-	{
-	    display_move(0, y_message);
-	    (void) clear_eol(msglen);
-	    msglen = 0;
+	    /* yes, free the current message */
+	    dprintf("i_message: timer expired\n");
+	    if (message_current != NULL)
+	    {
+		free(message_current);
+		message_current = NULL;
+	    }
+
+	    /* is there a new message to be displayed? */
+	    if (message_first != message_last)
+	    {
+		/* move index to next message */
+		if (++message_first == MAX_MESSAGES) message_first = 0;
+
+		/* make the next message the current one */
+		message_current = message_buf[message_first];
+
+		/* show it */
+		dprintf("i_message: showing \"%s\"\n", message_current);
+		display_move(0, y_message);
+		screen_standout(message_current);
+		i = strlen(message_current);
+
+		/* set the expiration timer */
+		message_time = *now;
+		message_time.tv_sec += MESSAGE_DISPLAY_TIME;
+
+		/* clear the rest of the line */
+		screen_cleareol(message_length - i);
+		putchar('\r');
+		message_length = i;
+	    }
+	    else
+	    {
+		/* just clear what was there before, if anything */
+		if (message_length > 0)
+		{
+		    display_move(0, y_message);
+		    screen_cleareol(message_length);
+		    putchar('\r');
+		    message_length = 0;
+		}
+	    }
 	}
     }
 }
 
 void
-u_message()
+u_message(struct timeval *now)
 
 {
-    i_message();
+    i_message(now);
 }
 
 static int header_length;
@@ -1239,15 +1443,16 @@ void
 i_header(char *text)
 
 {
+    int header_color = 0;
+
+#ifdef ENABLE_COLOR
+    header_color = color_test(header_cidx, 0);
+#endif
     header_length = strlen(text);
-    if (header_status == ON)
+    if (header_status)
     {
 	display_write(x_header, y_header, header_color, 1, text);
     }
-    else if (header_status == ERASE)
-    {
-	header_status = OFF;
-    }
 }
 
 /*ARGSUSED*/
@@ -1255,11 +1460,13 @@ void
 u_header(char *text)
 
 {
-    if (header_status == ERASE)
-    {
-	display_write(x_header, y_header, header_color, 1, "");
-	header_status = OFF;
-    }
+    int header_color = 0;
+
+#ifdef ENABLE_COLOR
+    header_color = color_test(header_cidx, 0);
+#endif
+    display_write(x_header, y_header, header_color, 1,
+		  header_status ? text : "");
 }
 
 /*
@@ -1280,14 +1487,31 @@ i_process(int line, char *thisline)
 }
 
 void
-u_process(int line, char *newline)
+u_process(int line, char *new_line)
 
 {
-    i_process(line, newline);
+    i_process(line, new_line);
 }
 
 void
-u_endscreen(int hi)
+i_endscreen()
+
+{
+    if (smart_terminal)
+    {
+	/* move the cursor to a pleasant place */
+	display_move(x_idlecursor, y_idlecursor);
+    }
+    else
+    {
+	/* separate this display from the next with some vertical room */
+	fputs("\n\n", stdout);
+    }
+    fflush(stdout);
+}
+
+void
+u_endscreen()
 
 {
     if (smart_terminal)
@@ -1296,8 +1520,8 @@ u_endscreen(int hi)
 	display_cte();
 
 	/* move the cursor to a pleasant place */
-	/* does this need to be a display_move??? */
-	Move_to(x_idlecursor, y_idlecursor);
+	display_move(x_idlecursor, y_idlecursor);
+	fflush(stdout);
     }
     else
     {
@@ -1310,90 +1534,229 @@ void
 display_header(int t)
 
 {
-    if (t)
-    {
-	header_status = ON;
-    }
-    else if (header_status == ON)
-    {
-	header_status = ERASE;
-    }
+    header_status = t != 0;
 }
 
 void
-new_message_v(int type, char *msgfmt, va_list ap)
+message_mark()
 
 {
-    register int i;
+    message_barrier = Yes;
+}
+
+void
+message_expire()
+
+{
+    message_time.tv_sec = 0;
+    message_time.tv_usec = 0;
+}
+
+void
+message_flush()
+
+{
+    message_first = message_last;
+    message_time.tv_sec = 0;
+    message_time.tv_usec = 0;
+}
+
+/*
+ * void new_message_v(char *msgfmt, va_list ap)
+ *
+ * Display a message in the message area.  This function takes a va_list for
+ * the arguments.  Safe to call before display_init.  This function only
+ * queues a message for display, and allowed for multiple messages to be
+ * queued.  The i_message function drains the queue and actually writes the
+ * messages on the display.
+ */
+
+
+void
+new_message_v(char *msgfmt, va_list ap)
+
+{
+    int i;
+    int empty;
+    char msg[MAX_COLS];
+
+    /* if message_barrier is active, remove all pending messages */
+    if (message_barrier)
+    {
+	message_flush();
+	message_barrier = No;
+    }
 
     /* first, format the message */
-    (void) vsnprintf(next_msg, sizeof(next_msg), msgfmt, ap);
+    (void) vsnprintf(msg, sizeof(msg), msgfmt, ap);
 
-    if (msglen > 0)
+    /* where in the buffer will it go? */
+    i = message_last + 1;
+    if (i >= MAX_MESSAGES) i = 0;
+
+    /* make sure the buffer is not full */
+    if (i != message_first)
     {
-	/* message there already -- can we clear it? */
-	if (!overstrike)
+	/* insert it in to message_buf */
+	message_buf[i] = strdup(msg);
+	dprintf("new_message_v: new message inserted in slot %d\n", i);
+
+	/* remember if the buffer is empty and set the index */
+	empty = message_last == message_first;
+	message_last = i;
+
+	/* is message_buf otherwise empty and have we started displaying? */
+	if (empty && !message_hold)
 	{
-	    /* yes -- write it and clear to end */
-	    i = strlen(next_msg);
-	    if ((type & MT_delayed) == 0)
-	    {
-		if ((type & MT_standout) != 0)
-		    standout(next_msg);
-		else
-		    fputs(next_msg, stdout);
-		(void) clear_eol(msglen - i);
-		msglen = i;
-		next_msg[0] = '\0';
-	    }
+	    /* we can display the message now */
+	    i_message(NULL);
 	}
     }
+}
+
+/*
+ * void new_message(int type, char *msgfmt, ...)
+ *
+ * Display a message in the message area.  It is safe to call this function
+ * before display_init.  Messages logged before the display is drawn will be
+ * held and displayed later.
+ */
+
+void
+new_message(char *msgfmt, ...)
+
+{
+    va_list ap;
+
+    va_start(ap, msgfmt);
+    new_message_v(msgfmt, ap);
+    va_end(ap);
+}
+
+/*
+ * void message_error(char *msgfmt, ...)
+ *
+ * Put an error message in the message area.  It is safe to call this function
+ * before display_init.  Messages logged before the display is drawn will be
+ * held and displayed later.
+ */
+
+void
+message_error(char *msgfmt, ...)
+
+{
+    va_list ap;
+
+    va_start(ap, msgfmt);
+    new_message_v(msgfmt, ap);
+    fflush(stdout);
+    va_end(ap);
+}
+
+/*
+ * void message_clear()
+ *
+ * Clear message area and flush all pending messages.
+ */
+
+void
+message_clear()
+
+{
+    /* remove any existing message */
+    if (message_current != NULL)
+    {
+	display_move(0, y_message);
+	screen_cleareol(message_length);
+	free(message_current);
+	message_current = 0;
+    }
+
+    /* flush all pending messages */
+    message_flush();
+}
+
+/*
+ * void message_prompt_v(int so, char *msgfmt, va_list ap)
+ *
+ * Place a prompt in the message area.  A prompt is different from a 
+ * message as follows: it is displayed immediately, overwriting any
+ * message that may already be there, it may be highlighted in standout
+ * mode (if "so" is true), the cursor is left to rest at the end of the
+ * prompt.  This call causes all pending messages to be flushed.
+ */
+
+void
+message_prompt_v(int so, char *msgfmt, va_list ap)
+
+{
+    char msg[MAX_COLS];
+    int i;
+
+    /* clear out the message buffer */
+    message_flush();
+
+    /* format the message */
+    i = vsnprintf(msg, sizeof(msg), msgfmt, ap);
+
+    /* this goes over any existing message */
+    display_move(0, y_message);
+
+    /* show the prompt */
+    if (so)
+    {
+	screen_standout(msg);
+    }
     else
     {
-	if ((type & MT_delayed) == 0)
-	{
-	    if ((type & MT_standout) != 0)
-		standout(next_msg);
-	    else
-		fputs(next_msg, stdout);
-	    msglen = strlen(next_msg);
-	    next_msg[0] = '\0';
-	}
+	fputs(msg, stdout);
     }
+
+    /* make it all visible */
+    fflush(stdout);
+
+    /* even though we dont keep a copy of the prompt, track its length */
+    message_length = i < MAX_COLS ? i : MAX_COLS;
 }
 
+/*
+ * void message_prompt(char *msgfmt, ...)
+ *
+ * Place a prompt in the message area (see message_prompt_v).
+ */
+
 void
-new_message(int type, char *msgfmt, ...)
+message_prompt(char *msgfmt, ...)
 
 {
     va_list ap;
 
     va_start(ap, msgfmt);
-    new_message_v(type, msgfmt, ap);
+    message_prompt_v(Yes, msgfmt, ap);
     va_end(ap);
 }
 
 void
-error_message(char *msgfmt, ...)
+message_prompt_plain(char *msgfmt, ...)
 
 {
     va_list ap;
 
     va_start(ap, msgfmt);
-    new_message_v(MT_standout | MT_delayed, msgfmt, ap);
+    message_prompt_v(No, msgfmt, ap);
     va_end(ap);
 }
 
-void
-clear_message()
-
-{
-    dprintf("clear_message: msglen = %d, x = %d, y = %d\n", msglen, curr_x, curr_y);
-    if (clear_eol(msglen) == 1)
-    {
-	putchar('\r');
-    }
-}
+/*
+ * int readline(char *buffer, int size, int numeric)
+ *
+ * Read a line of input from the terminal.  The line is placed in
+ * "buffer" not to exceed "size".  If "numeric" is true then the input
+ * can only consist of digits.  This routine handles all character
+ * editing while keeping the terminal in cbreak mode.  If "numeric"
+ * is true then the number entered is returned.  Otherwise the number
+ * of character read in to "buffer" is returned.
+ */
 
 int
 readline(char *buffer, int size, int numeric)
@@ -1402,7 +1765,6 @@ readline(char *buffer, int size, int numeric)
     register char *ptr = buffer;
     register char ch;
     register char cnt = 0;
-    register char maxcnt = 0;
 
     /* allow room for null terminator */
     size -= 1;
@@ -1419,12 +1781,6 @@ readline(char *buffer, int size, int numeric)
 	/* handle special editing characters */
 	if (ch == ch_kill)
 	{
-	    /* kill line -- account for overstriking */
-	    if (overstrike)
-	    {
-		msglen += maxcnt;
-	    }
-
 	    /* return null string */
 	    *buffer = '\0';
 	    putchar('\r');
@@ -1446,8 +1802,8 @@ readline(char *buffer, int size, int numeric)
 	    }
 	}
 	/* check for character validity and buffer overflow */
-	else if (cnt == size || (numeric && !isdigit(ch)) ||
-		!isprint(ch))
+	else if (cnt == size || (numeric && !isdigit((int)ch)) ||
+		!isprint((int)ch))
 	{
 	    /* not legal */
 	    putchar('\7');
@@ -1458,19 +1814,14 @@ readline(char *buffer, int size, int numeric)
 	    putchar(ch);
 	    ptr++;
 	    cnt++;
-	    if (cnt > maxcnt)
-	    {
-		maxcnt = cnt;
-	    }
 	}
     }
 
     /* all done -- null terminate the string */
     *ptr = '\0';
 
-    /* account for the extra characters in the message area */
-    /* (if terminal overstrikes, remember the furthest they went) */
-    msglen += overstrike ? maxcnt : cnt;
+    /* add response length to message_length */
+    message_length += cnt;
 
     /* return either inputted number or string length */
     putchar('\r');
@@ -1490,17 +1841,27 @@ display_pagerend()
 {
     char ch;
 
-    standout("Hit any key to continue: ");
+    screen_standout("Hit any key to continue: ");
     fflush(stdout);
     (void) read(0, &ch, 1);
 }
 
 void
-display_pager(char *data)
+display_pager(char *fmt, ...)
 
 {
+    va_list ap;
+
     int ch;
     char readch;
+    char buffer[MAX_COLS];
+    char *data;
+
+    /* format into buffer */
+    va_start(ap, fmt);
+    (void) vsnprintf(buffer, MAX_COLS, fmt, ap);
+    va_end(ap);
+    data = buffer;
 
     while ((ch = *data++) != '\0')
     {
@@ -1509,7 +1870,7 @@ display_pager(char *data)
 	{
 	    if (++curr_y >= screen_length - 1)
 	    {
-		standout("...More...");
+		screen_standout("...More...");
 		fflush(stdout);
 		(void) read(0, &readch, 1);
 		putchar('\r');
