@@ -1,3 +1,35 @@
+/*
+ * Copyright (c) 1984 through 2007, William LeFebvre
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *     * Redistributions of source code must retain the above copyright
+ * notice, this list of conditions and the following disclaimer.
+ * 
+ *     * Redistributions in binary form must reproduce the above
+ * copyright notice, this list of conditions and the following disclaimer
+ * in the documentation and/or other materials provided with the
+ * distribution.
+ * 
+ *     * Neither the name of William LeFebvre nor the names of other
+ * contributors may be used to endorse or promote products derived from
+ * this software without specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  * top - a top users display for Unix
  *
@@ -15,6 +47,7 @@
  * AUTHOR: Richard Henderson <rth@tamu.edu>
  * Order support added by Alexey Klimkin <kad@klon.tme.mcst.ru>
  * Ported to 2.4 by William LeFebvre
+ * Additions for 2.6 by William LeFebvre
  */
 
 #include "config.h"
@@ -44,8 +77,10 @@
 #endif
 
 #include "top.h"
+#include "hash.h"
 #include "machine.h"
 #include "utils.h"
+#include "username.h"
 
 #define PROCFS "/proc"
 extern char *myname;
@@ -57,12 +92,12 @@ struct top_proc
     pid_t pid;
     uid_t uid;
     char *name;
-    int pri, nice;
-    unsigned long size, rss;	/* in k */
+    int pri, nice, threads;
+    unsigned long size, rss, shared;	/* in k */
     int state;
     unsigned long time;
     unsigned long start_time;
-    double pcpu, wcpu;
+    double pcpu;
     struct top_proc *next;
 };
     
@@ -91,6 +126,16 @@ static char *cpustatenames[NCPUSTATES+1] =
 };
 static int show_iowait = 0;
 
+#define KERNELCTXT    0
+#define KERNELINTR    1
+#define KERNELNEWPROC 2
+#define NKERNELSTATS  3
+static char *kernelnames[NKERNELSTATS+1] =
+{
+    " ctxsw, ", " intr, ", " newproc",
+    NULL
+};
+
 #define MEMUSED    0
 #define MEMFREE    1
 #define MEMSHARED  2
@@ -114,7 +159,13 @@ static char *swapnames[NSWAPSTATS+1] =
 };
 
 static char fmt_header[] =
-"  PID X        PRI NICE  SIZE   RES STATE   TIME   WCPU    CPU COMMAND";
+"  PID X         THR PRI NICE  SIZE   RES STATE   TIME    CPU COMMAND";
+
+static char proc_header_thr[] =
+"  PID %-9s THR PRI NICE  SIZE   RES   SHR STATE   TIME    CPU COMMAND";
+
+static char proc_header_nothr[] =
+"  PID %-9s PRI NICE  SIZE   RES   SHR STATE   TIME    CPU COMMAND";
 
 /* these are names given to allowed sorting orders -- first is default */
 char *ordernames[] = 
@@ -145,31 +196,44 @@ static long cp_diff[NCPUSTATES];
 
 /* for calculating the exponential average */
 
-static struct timeval lasttime;
+static struct timeval lasttime = { 0, 0 };
+static struct timeval timediff = { 0, 0 };
+static long elapsed_msecs;
 
-/* these are for keeping track of processes */
+/* these are for keeping track of processes and tasks */
 
 #define HASH_SIZE	     (1003)
 #define INITIAL_ACTIVE_SIZE  (256)
 #define PROCBLOCK_SIZE       (32)
-static struct top_proc *ptable[HASH_SIZE];
+static hash_table *ptable;
+static hash_table *tasktable;
 static struct top_proc **pactive;
 static struct top_proc **nextactive;
 static unsigned int activesize = 0;
 static time_t boottime = -1;
+static int have_task = 0;
+
+/* these are counters that need to be track */
+static unsigned long last_ctxt = 0;
+static unsigned long last_intr = 0;
+static unsigned long last_newproc = 0;
 
 /* these are for passing data back to the machine independant portion */
 
 static int cpu_states[NCPUSTATES];
 static int process_states[NPROCSTATES];
+static int kernel_stats[NKERNELSTATS];
 static long memory_stats[NMEMSTATS];
 static long swap_stats[NSWAPSTATS];
 
-/* usefull macros */
+/* useful macros */
 #define bytetok(x)	(((x) + 512) >> 10)
 #define pagetok(x)	((x) << (PAGE_SHIFT - 10))
 #define HASH(x)		(((x) * 1686629713U) % HASH_SIZE)
 
+/* calculate a per-second rate using milliseconds */
+#define per_second(n, msec)   (((n) * 1000) / (msec))
+
 /*======================================================================*/
 
 static inline char *
@@ -293,9 +357,10 @@ machine_init(struct statics *statics)
 	int fd;
 	char buff[128];
 	char *p;
-	int cnt;
+	int cnt = 0;
 	unsigned long uptime;
 	struct timeval tv;
+	struct stat st;
 
 	/* get a boottime */
 	if ((fd = open("uptime", 0)) != -1)
@@ -318,7 +383,6 @@ machine_init(struct statics *statics)
 		{
 		    *p = '\0';
 		    p = buff;
-		    cnt = 0;
 		    while (*p != '\0')
 		    {
 			if (*p++ == ' ')
@@ -336,6 +400,13 @@ machine_init(struct statics *statics)
 	    /* we have iowait */
 	    show_iowait = 1;
 	}
+
+	/* see if we have task subdirs */
+	if (stat("self/task", &st) != -1 && S_ISDIR(st.st_mode))
+	{
+	    dprintf("we have task directories\n");
+	    have_task = 1;
+	}
     }
 
     /* if we aren't showing iowait, then we have to tweak cpustatenames */
@@ -347,19 +418,22 @@ machine_init(struct statics *statics)
     /* fill in the statics information */
     statics->procstate_names = procstatenames;
     statics->cpustate_names = cpustatenames;
+    statics->kernel_names = kernelnames;
     statics->memory_names = memorynames;
     statics->swap_names = swapnames;
     statics->order_names = ordernames;
     statics->boottime = boottime;
     statics->flags.fullcmds = 1;
     statics->flags.warmup = 1;
+    statics->flags.threads = 1;
 
     /* allocate needed space */
     pactive = (struct top_proc **)malloc(sizeof(struct top_proc *) * INITIAL_ACTIVE_SIZE);
     activesize = INITIAL_ACTIVE_SIZE;
 
-    /* make sure the hash table is empty */
-    memset(ptable, 0, HASH_SIZE * sizeof(struct top_proc *));
+    /* create process and task hashes */
+    ptable = hash_create(HASH_SIZE);
+    tasktable = hash_create(HASH_SIZE);
 
     /* all done! */
     return 0;
@@ -373,9 +447,18 @@ get_system_info(struct system_info *info)
     char buffer[4096+1];
     int fd, len;
     char *p;
+    struct timeval thistime;
+    unsigned long intr = 0;
+    unsigned long ctxt = 0;
+    unsigned long newproc = 0;
+
+    /* timestamp and time difference */
+    gettimeofday(&thistime, 0);
+    timersub(&thistime, &lasttime, &timediff);
+    elapsed_msecs = timediff.tv_sec * 1000 + timediff.tv_usec / 1000;
+    lasttime = thistime;
 
     /* get load averages */
-
     if ((fd = open("loadavg", O_RDONLY)) != -1)
     {
 	if ((len = read(fd, buffer, sizeof(buffer)-1)) > 0)
@@ -416,6 +499,37 @@ get_system_info(struct system_info *info)
 
 	    /* convert cp_time counts to percentages */
 	    percentages(NCPUSTATES, cpu_states, cp_time, cp_old, cp_diff);
+
+	    /* get the rest of it */
+	    p = strchr(p, '\n');
+	    while (p != NULL)
+	    {
+		p++;
+		if (strncmp(p, "intr ", 5) == 0)
+		{
+		    p = skip_token(p);
+		    intr = strtoul(p, &p, 10);
+		}
+		else if (strncmp(p, "ctxt ", 5) == 0)
+		{
+		    p = skip_token(p);
+		    ctxt = strtoul(p, &p, 10);
+		}
+		else if (strncmp(p, "processes ", 10) == 0)
+		{
+		    p = skip_token(p);
+		    newproc = strtoul(p, &p, 10);
+		}
+
+		p = strchr(p, '\n');
+	    }
+
+	    kernel_stats[KERNELINTR] = per_second(intr - last_intr, elapsed_msecs);
+	    kernel_stats[KERNELCTXT] = per_second(ctxt - last_ctxt, elapsed_msecs);
+	    kernel_stats[KERNELNEWPROC] = per_second(newproc - last_newproc, elapsed_msecs);
+	    last_intr = intr;
+	    last_ctxt = ctxt;
+	    last_newproc = newproc;
 	}
 	close(fd);
     }
@@ -525,16 +639,18 @@ get_system_info(struct system_info *info)
     info->cpustates = cpu_states;
     info->memory = memory_stats;
     info->swap = swap_stats;
+    info->kernel = kernel_stats;
 }
 
-
 static void
-read_one_proc_stat(pid_t pid, struct top_proc *proc, struct process_select *sel)
+read_one_proc_stat(pid_t pid, pid_t taskpid, struct top_proc *proc, struct process_select *sel)
 {
     char buffer[4096], *p, *q;
     int fd, len;
     int fullcmd;
 
+    dprintf("reading proc %d - %d\n", pid, taskpid);
+
     /* if anything goes wrong, we return with proc->state == 0 */
     proc->state = 0;
 
@@ -542,7 +658,14 @@ read_one_proc_stat(pid_t pid, struct top_proc *proc, struct process_select *sel)
     fullcmd = sel->fullcmd;
     if (fullcmd)
     {
-	sprintf(buffer, "%d/cmdline", pid);
+	if (taskpid == -1)
+	{
+	    sprintf(buffer, "%d/cmdline", pid);
+	}
+	else
+	{
+	    sprintf(buffer, "%d/task/%d/cmdline", pid, taskpid);
+	}
 	if ((fd = open(buffer, O_RDONLY)) != -1)
 	{
 	    /* read command line data */
@@ -565,8 +688,26 @@ read_one_proc_stat(pid_t pid, struct top_proc *proc, struct process_select *sel)
 	}
     }
 
+    /* grab the shared memory size */
+    sprintf(buffer, "%d/statm", pid);
+    fd = open(buffer, O_RDONLY);
+    len = read(fd, buffer, sizeof(buffer)-1);
+    close(fd);
+    buffer[len] = '\0';
+    p = buffer;
+    p = skip_token(p);		/* skip size */
+    p = skip_token(p);		/* skip resident */
+    proc->shared = pagetok(strtoul(p, &p, 10));
+
     /* grab the proc stat info in one go */
-    sprintf(buffer, "%d/stat", pid);
+    if (taskpid == -1)
+    {
+	sprintf(buffer, "%d/stat", pid);
+    }
+    else
+    {
+	sprintf(buffer, "%d/task/%d/stat", pid, taskpid);
+    }
 
     fd = open(buffer, O_RDONLY);
     len = read(fd, buffer, sizeof(buffer)-1);
@@ -628,8 +769,8 @@ read_one_proc_stat(pid_t pid, struct top_proc *proc, struct process_select *sel)
 
     proc->pri = strtol(p, &p, 10);		/* priority */
     proc->nice = strtol(p, &p, 10);		/* nice */
+    proc->threads = strtol(p, &p, 10);		/* threads */
 
-    p = skip_token(p);				/* skip timeout */
     p = skip_token(p);				/* skip it_real_val */
     proc->start_time = strtoul(p, &p, 10);	/* start_time */
 
@@ -653,76 +794,81 @@ read_one_proc_stat(pid_t pid, struct top_proc *proc, struct process_select *sel)
 
 }
 
+static int show_usernames;
+static int show_threads;
+
 
 caddr_t
 get_process_info(struct system_info *si,
 		 struct process_select *sel,
 		 int compare_index)
 {
-    struct timeval thistime;
-    double timediff, alpha, beta;
     struct top_proc *proc;
+    struct top_proc *taskproc;
     pid_t pid;
+    pid_t taskpid;
     unsigned long now;
     unsigned long elapsed;
-    int i;
-
-    /* calculate the time difference since our last check */
-    gettimeofday(&thistime, 0);
-    if (lasttime.tv_sec)
-    {
-	timediff = ((thistime.tv_sec - lasttime.tv_sec) +
-		    (thistime.tv_usec - lasttime.tv_usec) * 1e-6);
-    }
-    else
-    {
-	timediff = 0;
-    }
-    lasttime = thistime;
+    hash_item_pid *hi;
+    hash_pos pos;
 
     /* round current time to a second */
-    now = (unsigned long)thistime.tv_sec;
-    if (thistime.tv_usec >= 500000)
+    now = (unsigned long)lasttime.tv_sec;
+    if (lasttime.tv_usec >= 500000)
     {
 	now++;
     }
 
-    /* calculate constants for the exponental average */
-    if (timediff > 0.0 && timediff < 30.0)
+    /* calculate number of ticks since our last check */
+    elapsed = timediff.tv_sec * HZ + (timediff.tv_usec * HZ) / 1000000;
+    if (elapsed <= 0)
     {
-	alpha = 0.5 * (timediff / 30.0);
-	beta = 1.0 - alpha;
+	elapsed = 1;
     }
-    else
-    {
-	alpha = beta = 0.5;
-    }
-    timediff *= HZ;  /* convert to ticks */
+    dprintf("get_process_info: elapsed %d ticks\n", elapsed);
 
     /* mark all hash table entries as not seen */
-    for (i = 0; i < HASH_SIZE; ++i)
+    hi = hash_first_pid(ptable, &pos);
+    while (hi != NULL)
     {
-	for (proc = ptable[i]; proc; proc = proc->next)
-	{
-	    proc->state = 0;
-	}
+	((struct top_proc *)(hi->value))->state = 0;
+	hi = hash_next_pid(&pos);
+    }
+    /* mark all hash table entries as not seen */
+    hi = hash_first_pid(tasktable, &pos);
+    while (hi != NULL)
+    {
+	((struct top_proc *)(hi->value))->state = 0;
+	hi = hash_next_pid(&pos);
     }
 
     /* read the process information */
     {
 	DIR *dir = opendir(".");
+	DIR *taskdir;
 	struct dirent *ent;
+	struct dirent *taskent;
 	int total_procs = 0;
 	struct top_proc **active;
+	hash_item_pid *hi;
+	hash_pos pos;
+	char buffer[64];
 
 	int show_idle = sel->idle;
 	int show_uid = sel->uid != -1;
+	char *show_command = sel->command;
+
+	show_usernames = sel->usernames;
+	show_threads = sel->threads && have_task;
 
 	memset(process_states, 0, sizeof(process_states));
 
+	taskdir = NULL;
+	taskent = NULL;
+	taskpid = -1;
+
 	while ((ent = readdir(dir)) != NULL)
 	{
-	    struct top_proc *pp;
 	    unsigned long otime;
 
 	    if (!isdigit(ent->d_name[0]))
@@ -731,52 +877,97 @@ get_process_info(struct system_info *si,
 	    pid = atoi(ent->d_name);
 
 	    /* look up hash table entry */
-	    proc = pp = ptable[HASH(pid)];
-	    while (proc && proc->pid != pid)
-	    {
-		proc = proc->next;
-	    }
+	    proc = hash_lookup_pid(ptable, pid);
 
 	    /* if we came up empty, create a new entry */
 	    if (proc == NULL)
 	    {
 		proc = new_proc();
 		proc->pid = pid;
-		proc->next = pp;
-		ptable[HASH(pid)] = proc;
 		proc->time = 0;
-		proc->wcpu = 0;
+		hash_add_pid(ptable, pid, (void *)proc);
 	    }
 
+	    /* remember the previous cpu time */
 	    otime = proc->time;
 
-	    read_one_proc_stat(pid, proc, sel);
+	    /* get current data */
+	    read_one_proc_stat(pid, -1, proc, sel);
 
+	    /* continue on if this isn't really a process */
 	    if (proc->state == 0)
 		continue;
 
+	    /* reset linked list (for threads) */
+	    proc->next = NULL;
+
+	    /* accumulate process state data */
 	    total_procs++;
 	    process_states[proc->state]++;
 
-	    if (timediff > 0.0)
+	    /* calculate pcpu */
+	    if ((proc->pcpu = (proc->time - otime) / (double)elapsed) < 0.0001)
 	    {
-		if ((proc->pcpu = (proc->time - otime) / timediff) < 0.0001)
-		{
-		    proc->pcpu = 0;
-		}
-		proc->wcpu = proc->pcpu * alpha + proc->wcpu * beta;
+		proc->pcpu = 0;
 	    }
-	    else if ((elapsed = (now - boottime)*HZ - proc->start_time) > 0)
+
+	    /* if we have task subdirs and this process has more than
+	       one thread, collect data on each thread */
+	    if (have_task && proc->threads > 1)
 	    {
-		if ((proc->pcpu = (double)proc->time / (double)elapsed) < 0.0001)
+		snprintf(buffer, sizeof(buffer), "%d/task", pid);
+		if ((taskdir = opendir(buffer)) != NULL)
 		{
-		    proc->pcpu;
+		    while ((taskent = readdir(taskdir)) != NULL)
+		    {
+			if (!isdigit(taskent->d_name[0]))
+			    continue;
+
+			/* lookup entry in tasktable */
+			taskpid = atoi(taskent->d_name);
+			taskproc = hash_lookup_pid(tasktable, taskpid);
+
+			/* if we came up empty, create a new entry */
+			if (taskproc == NULL)
+			{
+			    taskproc = new_proc();
+			    taskproc->pid = taskpid;
+			    taskproc->time = 0;
+			    hash_add_pid(tasktable, taskpid, (void *)taskproc);
+			}
+
+			/* remember the previous cpu time */
+			otime = taskproc->time;
+
+			/* get current data */
+			read_one_proc_stat(pid, taskpid, taskproc, sel);
+
+			/* ignore if it isnt real */
+			if (taskproc->state == 0)
+			    continue;
+
+			/* when showing threads, add this to the accumulated
+			   process state data, but remember that the first
+			   thread is already accounted for */
+			if (show_threads && pid != taskpid)
+			{
+			    total_procs++;
+			    process_states[taskproc->state]++;
+			}
+
+			/* calculate pcpu */
+			if ((taskproc->pcpu = (taskproc->time - otime) /
+			     (double)elapsed) < 0.0)
+			{
+			    taskproc->pcpu = 0;
+			}
+
+			/* link this in to the proc's list */
+			taskproc->next = proc->next;
+			proc->next = taskproc;
+		    }
+		    closedir(taskdir);
 		}
-		proc->wcpu = proc->pcpu;
-	    }
-	    else
-	    {
-		proc->wcpu = proc->pcpu = 0.0;
 	    }
 	}
 	closedir(dir);
@@ -785,45 +976,51 @@ get_process_info(struct system_info *si,
 	if (activesize < total_procs)
 	{
 	    pactive = (struct top_proc **)realloc(pactive,
-						  sizeof(struct top_proc *) * total_procs);
+				  sizeof(struct top_proc *) * total_procs);
 	    activesize = total_procs;
 	}
 
 	/* set up the active procs and flush dead entries */
 	active = pactive;
-	for (i = 0; i < HASH_SIZE; i++)
+	hi = hash_first_pid(ptable, &pos);
+	while (hi != NULL)
 	{
-	    struct top_proc *last;
-	    struct top_proc *ptmp;
-
-	    last = NULL;
-	    proc = ptable[i];
-	    while (proc != NULL)
+	    proc = (struct top_proc *)(hi->value);
+	    if (proc->state == 0)
 	    {
-		if (proc->state == 0)
+		/* dead entry */
+		hash_remove_pos_pid(&pos);
+		free_proc(proc);
+	    }
+	    else
+	    {
+		/* check to see if it qualifies as active */
+		if ((show_idle || proc->state == 1 || proc->pcpu) &&
+		    (!show_uid || proc->uid == sel->uid) &&
+		    (show_command == NULL ||
+		     strstr(proc->name, show_command) != NULL))
 		{
-		    ptmp = proc;
-		    if (last)
+		    /* are we showing threads and does this proc have any? */
+		    if (show_threads && proc->threads > 1 && proc->next != NULL)
 		    {
-			proc = last->next = proc->next;
+			/* then add just the thread info -- the main process
+			   info is included in the list */
+			proc = proc->next;
+			while (proc != NULL)
+			{
+			    *active++ = proc;
+			    proc = proc->next;
+			}
 		    }
 		    else
 		    {
-			proc = ptable[i] = proc->next;
-		    }
-		    free_proc(ptmp);
-		}
-		else
-		{
-		    if ((show_idle || proc->state == 1 || proc->pcpu) &&
-			(!show_uid || proc->uid == sel->uid))
-		    {
+			/* add the process */
 			*active++ = proc;
-			last = proc;
 		    }
-		    proc = proc->next;
 		}
 	    }
+
+	    hi = hash_next_pid(&pos);
 	}
 
 	si->p_active = active - pactive;
@@ -855,6 +1052,21 @@ format_header(char *uname_field)
     return fmt_header;
 }
 
+static char p_header[MAX_COLS];
+
+char *
+format_process_header(struct process_select *sel, caddr_t handle, int count)
+
+{
+    char *h;
+
+    h = sel->threads ? proc_header_nothr : proc_header_thr;
+
+    snprintf(p_header, MAX_COLS, h, sel->usernames ? "USERNAME" : "UID");
+
+    return p_header;
+}
+
 
 char *
 format_next_process(caddr_t handle, char *(*get_userid)(int))
@@ -862,20 +1074,43 @@ format_next_process(caddr_t handle, char *(*get_userid)(int))
 {
     static char fmt[MAX_COLS];	/* static area where result is built */
     struct top_proc *p = *nextactive++;
+    char *userbuf;
 
-    snprintf(fmt, sizeof(fmt),
-	    "%5d %-8.8s %3d %4d %5s %5s %-5s %6s %5.2f%% %5.2f%% %s",
-	    p->pid,
-	    (*get_userid)(p->uid),
-	    p->pri < -99 ? -99 : p->pri,
-	    p->nice,
-	    format_k(p->size),
-	    format_k(p->rss),
-	    state_abbrev[p->state],
-	    format_time(p->time / HZ),
-	    p->wcpu * 100.0,
-	    p->pcpu * 100.0,
-	    p->name);
+    userbuf = show_usernames ? username(p->uid) : itoa_w(p->uid, 7);
+
+    if (show_threads)
+    {
+	snprintf(fmt, sizeof(fmt),
+		 "%5d %-8.8s  %3d %4d %5s %5s %5s %-5s %6s %5s%% %s",
+		 p->pid,
+		 userbuf,
+		 p->pri < -99 ? -99 : p->pri,
+		 p->nice,
+		 format_k(p->size),
+		 format_k(p->rss),
+		 format_k(p->shared),
+		 state_abbrev[p->state],
+		 format_time(p->time / HZ),
+		 format_percent(p->pcpu * 100.0),
+		 p->name);
+    }
+    else
+    {
+	snprintf(fmt, sizeof(fmt),
+		 "%5d %-8.8s %4d %3d %4d %5s %5s %5s %-5s %6s %5s%% %s",
+		 p->pid,
+		 userbuf,
+		 p->threads <= 9999 ? p->threads : 9999,
+		 p->pri < -99 ? -99 : p->pri,
+		 p->nice,
+		 format_k(p->size),
+		 format_k(p->rss),
+		 format_k(p->shared),
+		 state_abbrev[p->state],
+		 format_time(p->time / HZ),
+		 format_percent(p->pcpu * 100.0),
+		 p->name);
+    }
 
     /* return the result */
     return (fmt);
