@@ -10,11 +10,6 @@
  * through 5.9 (otherwise known as Solaris 9).  It has not been
  * tested on SunOS 5.10.
  *
- * LIBS: -lelf -lkvm -lkstat
- *
- * CFLAGS: -DHAVE_GETOPT -DORDER -DHAVE_STRERROR -DUSE_SIZE_T
- *
- *
  * AUTHORS:      Torsten Kasch 		<torsten@techfak.uni-bielefeld.de>
  *               Robert Boucher		<boucher@sofkin.ca>
  * CONTRIBUTORS: Marc Cohen 		<marc@aai.com>
@@ -27,7 +22,7 @@
 
 #define _KMEMUSER
 
-#include "config.h"
+#include "os.h"
 
 #if (OSREV == 551)
 #undef OSREV
@@ -67,6 +62,7 @@
 
 #include "top.h"
 #include "machine.h"
+#include <limits.h>
 #include <stdio.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -100,6 +96,7 @@
 #include <math.h>
 #include <utmpx.h>
 #include "utils.h"
+#include "hash.h"
 
 #if OSREV >= 53
 #define USE_KSTAT
@@ -212,9 +209,27 @@ struct oldproc
     pid_t oldpid;
     double oldtime;
     double oldpct;
+    int seen;
+};
+
+hash_table *prochash;
+
+/*
+ * Structure for tracking per-cpu information
+ */
+struct cpustats
+{
+    unsigned int states[CPUSTATES];
+    uint_t pswitch;
+    uint_t trap;
+    uint_t intr;
+    uint_t syscall;
+    uint_t sysfork;
+    uint_t sysvfork;
+    uint_t pfault;
+    uint_t pgin;
+    uint_t pgout;
 };
-int oldprocs;			/* size of table */
-#define HASH(x) ((x << 1) % oldprocs)
 
 /*
  * GCC assumes that all doubles are aligned.  Unfortunately it
@@ -273,6 +288,20 @@ char *memorynames[] =
 #define MEMORY_TOTALSWAP 2
 #define MEMORY_FREESWAP  3
 
+/* these are for detailing kernel statistics */
+int kernel_stats[8];
+char *kernelnames[] =
+{" ctxsw, ", " trap, ", " intr, ", " syscall, ", " fork, ",
+ " flt, ", " pgin, ", " pgout, ", NULL};
+#define KERNEL_CSWITCH 0
+#define KERNEL_TRAP 1
+#define KERNEL_INTR 2
+#define KERNEL_SYSCALL 3
+#define KERNEL_FORK 4
+#define KERNEL_PFAULT 5
+#define KERNEL_PGIN 6
+#define KERNEL_PGOUT 7
+
 /* these are names given to allowed sorting orders -- first is default */
 char *ordernames[] = 
 {"cpu", "size", "res", "time", NULL};
@@ -302,7 +331,6 @@ static int cpucount;
 /* these are for keeping track of the proc array */
 static struct prpsinfo *pbase;
 static struct prpsinfo **pref;
-static struct oldproc *oldbase;
 
 /* pagetok function is really a pointer to an appropriate function */
 static int pageshift;
@@ -311,12 +339,9 @@ static long (*p_pagetok) ();
 
 /* useful externals */
 extern char *myname;
-extern int check_nlist ();
-extern int gettimeofday ();
 extern void perror ();
 extern void getptable ();
 extern void quit ();
-extern int nlist ();
 
 /* p_pagetok points to one of the following, depending on which
    direction data has to be shifted: */
@@ -378,159 +403,35 @@ getkval (unsigned long offset,
 
 }
 
+/*
+ * check_nlist(nlst) - checks the nlist to see if any symbols were not
+ *		found.  For every symbol that was not found, a one-line
+ *		message is printed to stderr.  The routine returns the
+ *		number of symbols NOT found.
+ */
 int
-machine_init (struct statics *statics)
+check_nlist (register struct nlist *nlst)
 {
-    struct utmpx ut;
-    struct utmpx *up;
-    int i;
-    char *p;
-#ifndef USE_KSTAT
-    int offset;
-#endif
+    register int i;
 
-    /* There's a buffer overflow bug in curses that can be exploited when
-       we run as root.  By making sure that TERMINFO is set to something
-       this bug is avoided.  This code thanks to Casper */
-    if ((p = getenv("TERMINFO")) == NULL || *p == '\0')
+    /* check to see if we got ALL the symbols we requested */
+    /* this will write one line to stderr for every symbol not found */
+
+    i = 0;
+    while (nlst->n_name != NULL)
     {
-        putenv("TERMINFO=/usr/share/lib/terminfo/");
+	if (nlst->n_type == 0)
+	{
+	    /* this one wasn't found */
+	    fprintf (stderr, "kernel: no symbol named `%s'\n", nlst->n_name);
+	    i = 1;
+	}
+	nlst++;
     }
-
-    /* perform the kvm_open - suppress error here */
-    if ((kd = kvm_open (NULL, NULL, NULL, O_RDONLY, NULL)) == NULL)
-    {
-	/* save the error message: we may need it later */
-	p = strerror(errno);
-    }
-    dprintf("kvm_open: fd %d\n", kd);
-
-    /*
-     * turn off super group/user privs - but beware; we might
-     * want the privs back later and we still have a fd to
-     * /dev/kmem open so we can't use setgid()/setuid() as that
-     * would allow a debugger to attach to this process. CD
-     */
-    setegid(getgid());
-    seteuid(getuid()); /* super user not needed for NEW_PROC */
-
-#ifdef USE_KSTAT
-    if ((kc = kstat_open()) == NULL)
-    {
-	fprintf(stderr, "Unable to open kstat.\n");
-	return(-1);
-    }
-    kcid = kc->kc_chain_id;
-    dprintf("kstat_open: chain %d\n", kcid);
-#endif
-
-    /* fill in the statics information */
-    statics->procstate_names = procstatenames;
-    statics->cpustate_names = cpustatenames;
-    statics->memory_names = memorynames;
-    statics->order_names = ordernames;
-    statics->flags.fullcmds = 1;
-    statics->flags.warmup = 1;
-
-    /* get boot time */
-    ut.ut_type = BOOT_TIME;
-    if ((up = getutxid(&ut)) != NULL)
-    {
-	statics->boottime = up->ut_tv.tv_sec;
-    }
-    endutxent();
-
-    /* if the kvm_open succeeded, get the nlist */
-    if (kd)
-    {
-	if (kvm_nlist (kd, nlst) < 0)
-        {
-	    perror ("kvm_nlist");
-	    return (-1);
-        }
-	if (check_nlist (nlst) != 0)
-	    return (-1);
-    }
-#ifndef USE_KSTAT
-    /* if KSTAT is not available to us and we can't open /dev/kmem,
-       this is a serious problem.
-    */
-    else
-    {
-	/* Print the error message here */
-	(void) fprintf(stderr, "kvm_open: %s\n", p);
-	return (-1);
-    }
-#endif
-
-    /* stash away certain offsets for later use */
-    mpid_offset = nlst[X_MPID].n_value;
-    nproc_offset = nlst[X_NPROC].n_value;
-    avenrun_offset = nlst[X_AVENRUN].n_value;
-    anoninfo_offset = nlst[X_ANONINFO].n_value;
-    freemem_offset = nlst[X_FREEMEM].n_value;
-    maxmem_offset = nlst[X_MAXMEM].n_value;
-
-
-#ifndef USE_KSTAT
-    (void) getkval (nlst[X_NCPUS].n_value, (int *) (&cpucount),
-		    sizeof (cpucount), "ncpus");
-
-    cpu_offset = (unsigned long *) malloc (cpucount * sizeof (unsigned long));
-    for (i = offset = 0; i < cpucount; offset += sizeof(unsigned long)) {
-        (void) getkval (nlst[X_CPU].n_value + offset,
-                        (int *)(&cpu_offset[i]), sizeof (unsigned long),
-                        nlst[X_CPU].n_name );
-        if (cpu_offset[i] != 0)
-            i++;
-    }
-#endif
-
-    /* calculate pageshift value */
-    i = sysconf(_SC_PAGESIZE);
-    pageshift = 0;
-    while ((i >>= 1) > 0)
-    {
-	pageshift++;
-    }
-
-    /* calculate an amount to shift to K values */
-    /* remember that log base 2 of 1024 is 10 (i.e.: 2^10 = 1024) */
-    pageshift -= 10;
-
-    /* now determine which pageshift function is appropriate for the 
-       result (have to because x << y is undefined for y < 0) */
-    if (pageshift > 0)
-    {
-	/* this is the most likely */
-	p_pagetok = pagetok_left;
-    }
-    else if (pageshift == 0)
-    {
-	p_pagetok = pagetok_none;
-    }
-    else
-    {
-	p_pagetok = pagetok_right;
-	pageshift = -pageshift;
-    }
-
-    if (!(procdir = opendir (PROCFS)))
-    {
-	(void) fprintf (stderr, "Unable to open %s\n", PROCFS);
-	return (-1);
-    }
-
-    if (chdir (PROCFS))
-    {				/* handy for later on when we're reading it */
-	(void) fprintf (stderr, "Unable to chdir to %s\n", PROCFS);
-	return (-1);
-    }
-
-    /* all done! */
-    return (0);
+    return (i);
 }
 
+
 char *
 format_header (register char *uname_field)
 {
@@ -549,6 +450,7 @@ long
 kstat_data_value_l(kstat_named_t *kn)
 
 {
+#ifdef KSTAT_DATA_UINT32
     switch(kn->data_type)
     {
     case KSTAT_DATA_INT32:
@@ -561,6 +463,9 @@ kstat_data_value_l(kstat_named_t *kn)
 	return ((long)(kn->value.ui64));
     }
     return 0;
+#else
+    return ((long)(kn->value.ui32));
+#endif
 }
 
 int
@@ -630,7 +535,7 @@ kstat_safe_namematch(int num, kstat_t **ksparg, char *name, void *buf, int size)
     char *cbuf;
 
     dprintf("kstat_safe_namematch(%d, %08x, %s, %08x, %d)\n",
-	    num, ksp, name, buf, size);
+	    num, ksparg, name, buf, size);
 
     namelen = strlen(name);
 
@@ -755,18 +660,18 @@ get_nproc()
     {
 	ret = (int)(kn->value.ui32);
     }
-
-    return ret;
 #else
     int ret;
 
     (void) getkval (nproc_offset, (int *) (&ret), sizeof (ret), "nproc");
-    return ret;
 #endif
+
+    dprintf("get_nproc returns %d\n", ret);
+    return ret;
 }
 
-unsigned int
-(*get_cpustats(int *cnt, unsigned int (*cp_stats)[CPUSTATES]))[CPUSTATES]
+struct cpustats *
+get_cpustats(int *cnt, struct cpustats *cpustats)
 
 {
 #ifdef USE_KSTAT
@@ -774,10 +679,10 @@ unsigned int
     static cpu_stat_t *cpu_stat = NULL;
     static unsigned int nelems = 0;
     cpu_stat_t *cpu_stat_p;
-    int i, ret, cpu_num;
-    unsigned int (*cp_stats_p)[CPUSTATES];
+    int i, cpu_num;
+    struct cpustats *cpustats_p;
 
-    dprintf("get_cpustats(%d -> %d, %08x)\n", cnt, *cnt, cp_stats);
+    dprintf("get_cpustats(%d -> %d, %08x)\n", cnt, *cnt, cpustats);
 
     while (nelems > 0 ?
 	   (cpu_num = kstat_safe_namematch(nelems,
@@ -808,12 +713,12 @@ unsigned int
 	/* yes, so realloc their array, too */
 	dprintf("realloc array from %d to %d\n", *cnt, cpu_num);
 	*cnt = cpu_num;
-	cp_stats = (unsigned int (*)[CPUSTATES])realloc(cp_stats,
-			 cpu_num * sizeof(unsigned int) * CPUSTATES);
+	cpustats = (struct cpustats *)realloc(cpustats,
+					      cpu_num * sizeof(struct cpustats));
     }
 
     cpu_stat_p = cpu_stat;
-    cp_stats_p = cp_stats;
+    cpustats_p = cpustats;
     for (i = 0; i < cpu_num; i++)
     {
 	dprintf("cpu %d %08x: idle %u, user %u, syscall %u\n", i, cpu_stat_p,
@@ -821,21 +726,31 @@ unsigned int
 		cpu_stat_p->cpu_sysinfo.cpu[1],
 		cpu_stat_p->cpu_sysinfo.syscall);
 
-	(*cp_stats_p)[CPU_IDLE] = cpu_stat_p->cpu_sysinfo.cpu[CPU_IDLE];
-	(*cp_stats_p)[CPU_USER] = cpu_stat_p->cpu_sysinfo.cpu[CPU_USER];
-	(*cp_stats_p)[CPU_KERNEL] = cpu_stat_p->cpu_sysinfo.cpu[CPU_KERNEL];
-	(*cp_stats_p)[CPUSTATE_IOWAIT] = cpu_stat_p->cpu_sysinfo.wait[W_IO] +
+	cpustats_p->states[CPU_IDLE] = cpu_stat_p->cpu_sysinfo.cpu[CPU_IDLE];
+	cpustats_p->states[CPU_USER] = cpu_stat_p->cpu_sysinfo.cpu[CPU_USER];
+	cpustats_p->states[CPU_KERNEL] = cpu_stat_p->cpu_sysinfo.cpu[CPU_KERNEL];
+	cpustats_p->states[CPUSTATE_IOWAIT] = cpu_stat_p->cpu_sysinfo.wait[W_IO] +
 	    cpu_stat_p->cpu_sysinfo.wait[W_PIO];
-	(*cp_stats_p)[CPUSTATE_SWAP] = cpu_stat_p->cpu_sysinfo.wait[W_SWAP];
-	cp_stats_p++;
+	cpustats_p->states[CPUSTATE_SWAP] = cpu_stat_p->cpu_sysinfo.wait[W_SWAP];
+	cpustats_p->pswitch = cpu_stat_p->cpu_sysinfo.pswitch;
+	cpustats_p->trap = cpu_stat_p->cpu_sysinfo.trap;
+	cpustats_p->intr = cpu_stat_p->cpu_sysinfo.intr;
+	cpustats_p->syscall = cpu_stat_p->cpu_sysinfo.syscall;
+	cpustats_p->sysfork = cpu_stat_p->cpu_sysinfo.sysfork;
+	cpustats_p->sysvfork = cpu_stat_p->cpu_sysinfo.sysvfork;
+	cpustats_p->pfault = cpu_stat_p->cpu_vminfo.hat_fault +
+	    cpu_stat_p->cpu_vminfo.as_fault;
+	cpustats_p->pgin = cpu_stat_p->cpu_vminfo.pgin;
+	cpustats_p->pgout = cpu_stat_p->cpu_vminfo.pgout;
+	cpustats_p++;
 	cpu_stat_p++;
     }
 
     cpucount = cpu_num;
 
-    dprintf("get_cpustats sees %d cpus and returns %08x\n", cpucount, cp_stats);
+    dprintf("get_cpustats sees %d cpus and returns %08x\n", cpucount, cpustats);
 
-    return (cp_stats);
+    return (cpustats);
 #else /* !USE_KSTAT */
     int i;
     struct cpu cpu;
@@ -1006,41 +921,215 @@ get_swapinfo(long *total, long *fr)
     return (0);
 }
 
+int
+machine_init (struct statics *statics)
+{
+    struct utmpx ut;
+    struct utmpx *up;
+    int i;
+    char *p;
+#ifndef USE_KSTAT
+    int offset;
+#endif
+
+    /* There's a buffer overflow bug in curses that can be exploited when
+       we run as root.  By making sure that TERMINFO is set to something
+       this bug is avoided.  This code thanks to Casper */
+    if ((p = getenv("TERMINFO")) == NULL || *p == '\0')
+    {
+        putenv("TERMINFO=/usr/share/lib/terminfo/");
+    }
+
+    /* perform the kvm_open - suppress error here */
+    if ((kd = kvm_open (NULL, NULL, NULL, O_RDONLY, NULL)) == NULL)
+    {
+	/* save the error message: we may need it later */
+	p = strerror(errno);
+    }
+    dprintf("kvm_open: fd %d\n", kd);
+
+    /*
+     * turn off super group/user privs - but beware; we might
+     * want the privs back later and we still have a fd to
+     * /dev/kmem open so we can't use setgid()/setuid() as that
+     * would allow a debugger to attach to this process. CD
+     */
+    setegid(getgid());
+    seteuid(getuid()); /* super user not needed for NEW_PROC */
+
+#ifdef USE_KSTAT
+    /* open kstat */
+    if ((kc = kstat_open()) == NULL)
+    {
+	fprintf(stderr, "Unable to open kstat.\n");
+	return(-1);
+    }
+    kcid = kc->kc_chain_id;
+    dprintf("kstat_open: chain %d\n", kcid);
+#endif
+
+    /* fill in the statics information */
+    statics->procstate_names = procstatenames;
+    statics->cpustate_names = cpustatenames;
+    statics->memory_names = memorynames;
+    statics->kernel_names = kernelnames;
+    statics->order_names = ordernames;
+    statics->flags.fullcmds = 1;
+    statics->flags.warmup = 1;
+
+    /* get boot time */
+    ut.ut_type = BOOT_TIME;
+    if ((up = getutxid(&ut)) != NULL)
+    {
+	statics->boottime = up->ut_tv.tv_sec;
+    }
+    endutxent();
+
+    /* if the kvm_open succeeded, get the nlist */
+    if (kd)
+    {
+	if (kvm_nlist (kd, nlst) < 0)
+        {
+	    perror ("kvm_nlist");
+	    return (-1);
+        }
+	if (check_nlist (nlst) != 0)
+	    return (-1);
+    }
+#ifndef USE_KSTAT
+    /* if KSTAT is not available to us and we can't open /dev/kmem,
+       this is a serious problem.
+    */
+    else
+    {
+	/* Print the error message here */
+	(void) fprintf(stderr, "kvm_open: %s\n", p);
+	return (-1);
+    }
+#endif
+
+    /* stash away certain offsets for later use */
+    mpid_offset = nlst[X_MPID].n_value;
+    nproc_offset = nlst[X_NPROC].n_value;
+    avenrun_offset = nlst[X_AVENRUN].n_value;
+    anoninfo_offset = nlst[X_ANONINFO].n_value;
+    freemem_offset = nlst[X_FREEMEM].n_value;
+    maxmem_offset = nlst[X_MAXMEM].n_value;
+
+#ifndef USE_KSTAT
+    (void) getkval (nlst[X_NCPUS].n_value, (int *) (&cpucount),
+		    sizeof (cpucount), "ncpus");
+
+    cpu_offset = (unsigned long *) malloc (cpucount * sizeof (unsigned long));
+    for (i = offset = 0; i < cpucount; offset += sizeof(unsigned long)) {
+        (void) getkval (nlst[X_CPU].n_value + offset,
+                        (int *)(&cpu_offset[i]), sizeof (unsigned long),
+                        nlst[X_CPU].n_name );
+        if (cpu_offset[i] != 0)
+            i++;
+    }
+#endif
+
+    /* we need to get the current nproc */
+#ifdef USE_KSTAT
+    /* get_nproc assumes that the chain has already been retrieved,
+       so we need to do that here */
+    kstat_safe_retrieve(&ks_system_misc, "unix", 0, "system_misc", NULL);
+#endif
+    nproc = get_nproc();
+    dprintf("machine_init: nproc=%d\n", nproc);
+
+    /* hash table for remembering procs sized based on current nproc*/
+    prochash = hash_create(nproc > 100 ? nproc * 2 + 1 : 521);
+
+    /* calculate pageshift value */
+    i = sysconf(_SC_PAGESIZE);
+    pageshift = 0;
+    while ((i >>= 1) > 0)
+    {
+	pageshift++;
+    }
+
+    /* calculate an amount to shift to K values */
+    /* remember that log base 2 of 1024 is 10 (i.e.: 2^10 = 1024) */
+    pageshift -= 10;
+
+    /* now determine which pageshift function is appropriate for the 
+       result (have to because x << y is undefined for y < 0) */
+    if (pageshift > 0)
+    {
+	/* this is the most likely */
+	p_pagetok = pagetok_left;
+    }
+    else if (pageshift == 0)
+    {
+	p_pagetok = pagetok_none;
+    }
+    else
+    {
+	p_pagetok = pagetok_right;
+	pageshift = -pageshift;
+    }
+
+    if (!(procdir = opendir (PROCFS)))
+    {
+	(void) fprintf (stderr, "Unable to open %s\n", PROCFS);
+	return (-1);
+    }
+
+    if (chdir (PROCFS))
+    {				/* handy for later on when we're reading it */
+	(void) fprintf (stderr, "Unable to chdir to %s\n", PROCFS);
+	return (-1);
+    }
+
+    /* all done! */
+    return (0);
+}
+
 void
 get_system_info (struct system_info *si)
 {
     int avenrun[3];
-    static long freemem;
-    static long maxmem;
-    static int swap_total;
-    static int swap_free;
 
     static long cp_time[CPUSTATES];
     static long cp_old[CPUSTATES];
     static long cp_diff[CPUSTATES];
-    static unsigned int (*cp_stats)[CPUSTATES] = NULL;
-    static int cpus;
+    static struct cpustats *cpustats = NULL;
+    static struct cpustats sum_current;
+    static struct cpustats sum_old;
+    static int cpus = 0;
     register int j, i;
 
+    /* remember the old values and zero out the current */
+    memcpy(&sum_old, &sum_current, sizeof(sum_current));
+    memset(&sum_current, 0, sizeof(sum_current));
+
     /* get important information */
     get_avenrun(avenrun);
 
     /* get the cpu statistics arrays */
-    cp_stats = get_cpustats(&cpus, cp_stats);
+    cpustats = get_cpustats(&cpus, cpustats);
 
     /* zero the cp_time array */
-    for (j = 0; j < CPUSTATES; j++)
-    {
-	cp_time[j] = 0;
-    }
+    memset(cp_time, 0, sizeof(cp_time));
 
-    /* sum stats in to a single array */
+    /* sum stats in to a single array and a single structure */
     for (i = 0; i < cpus; i++)
     {
 	for (j = 0; j < CPUSTATES; j++)
 	{
-	    cp_time[j] += cp_stats[i][j];
+	    cp_time[j] += cpustats[i].states[j];
 	}
+	sum_current.pswitch += cpustats[i].pswitch;
+	sum_current.trap += cpustats[i].trap;
+	sum_current.intr += cpustats[i].intr;
+	sum_current.syscall += cpustats[i].syscall;
+	sum_current.sysfork += cpustats[i].sysfork;
+	sum_current.sysvfork += cpustats[i].sysvfork;
+	sum_current.pfault += cpustats[i].pfault;
+	sum_current.pgin += cpustats[i].pgin;
+	sum_current.pgout += cpustats[i].pgout;
     }
 
     /* convert cp_time counts to percentages */
@@ -1069,10 +1158,23 @@ get_system_info (struct system_info *si)
     {
 	memory_stats[MEMORY_TOTALSWAP] = memory_stats[MEMORY_FREESWAP] = -1;
     }
-  
+
+    /* get kernel data */
+    kernel_stats[KERNEL_CSWITCH] = diff_per_second(sum_current.pswitch, sum_old.pswitch);
+    kernel_stats[KERNEL_TRAP] = diff_per_second(sum_current.trap, sum_old.trap);
+    kernel_stats[KERNEL_INTR] = diff_per_second(sum_current.intr, sum_old.intr);
+    kernel_stats[KERNEL_SYSCALL] = diff_per_second(sum_current.syscall, sum_old.syscall);
+    kernel_stats[KERNEL_FORK] = diff_per_second(sum_current.sysfork + sum_current.sysvfork,
+						sum_old.sysfork + sum_old.sysvfork);
+    kernel_stats[KERNEL_PFAULT] = diff_per_second(sum_current.pfault, sum_old.pfault);
+    kernel_stats[KERNEL_PGIN] = pagetok(diff_per_second(sum_current.pgin, sum_old.pgin));
+    kernel_stats[KERNEL_PGOUT] = pagetok(diff_per_second(sum_current.pgout, sum_old.pgout));
+
+
     /* set arrays and strings */
     si->cpustates = cpu_states;
     si->memory = memory_stats;
+    si->kernel = kernel_stats;
 
     dprintf("get_system_info returns\n");
 }
@@ -1096,6 +1198,7 @@ get_process_info (
     int show_idle;
     int show_system;
     int show_uid;
+    char *show_command;
 
     /* allocate enough space for twice our current needs */
     nproc = get_nproc();
@@ -1115,6 +1218,7 @@ get_process_info (
     show_system = sel->system;
     show_uid = sel->uid != -1;
     show_fullcmd = sel->fullcmd;
+    show_command = sel->command;
 
     /* count up process states and get pointers to interesting procs */
     total_procs = 0;
@@ -1135,10 +1239,12 @@ get_process_info (
 	    (show_system || ((pp->pr_flag & SSYS) == 0)))
 	{
 	    total_procs++;
-	    process_states[pp->pr_state]++;
+	    process_states[(int)pp->pr_state]++;
 	    if ((!ZOMBIE(pp)) &&
 		(show_idle || percent_cpu (pp) || (pp->pr_state == SRUN) || (pp->pr_state == SONPROC)) &&
-		(!show_uid || pp->pr_uid == (uid_t) sel->uid))
+		(!show_uid || pp->pr_uid == (uid_t) sel->uid) &&
+		(show_command == NULL ||
+		 strstr(pp->pr_fname, show_command) != NULL))
 	    {
 		*prefp++ = pp;
 		active_procs++;
@@ -1146,6 +1252,8 @@ get_process_info (
 	}
     }
 
+    dprintf("total_procs %d, active_procs %d\n", total_procs, active_procs);
+
     /* if requested, sort the "interesting" processes */
     qsort ((char *) pref, active_procs, sizeof (struct prpsinfo *),
 	   proc_compares[compare_index]);
@@ -1201,7 +1309,7 @@ format_next_process (
 	   pp->pr_nice - NZERO,
 	   format_k(SIZE_K(pp)),
 	   format_k(RSS_K(pp)),
-	   *sb ? sb : state_abbrev[pp->pr_state],
+	   *sb ? sb : state_abbrev[(int)pp->pr_state],
 	   format_time(cputime),
 	   format_percent(pctcpu),
 	   printable(show_fullcmd ? pp->pr_psargs : pp->pr_fname));
@@ -1210,35 +1318,6 @@ format_next_process (
   return (fmt);
 }
 
-/*
- * check_nlist(nlst) - checks the nlist to see if any symbols were not
- *		found.  For every symbol that was not found, a one-line
- *		message is printed to stderr.  The routine returns the
- *		number of symbols NOT found.
- */
-int
-check_nlist (register struct nlist *nlst)
-{
-    register int i;
-
-    /* check to see if we got ALL the symbols we requested */
-    /* this will write one line to stderr for every symbol not found */
-
-    i = 0;
-    while (nlst->n_name != NULL)
-    {
-	if (nlst->n_type == 0)
-	{
-	    /* this one wasn't found */
-	    fprintf (stderr, "kernel: no symbol named `%s'\n", nlst->n_name);
-	    i = 1;
-	}
-	nlst++;
-    }
-    return (i);
-}
-
-
 /* comparison routines for qsort */
 
 /*
@@ -1261,8 +1340,8 @@ check_nlist (register struct nlist *nlst)
 #define ORDERKEY_PCTCPU  if (dresult = percent_cpu (p2) - percent_cpu (p1),\
 			     (result = dresult > 0.0 ? 1 : dresult < 0.0 ? -1 : 0) == 0)
 #define ORDERKEY_CPTICKS if ((result = p2->pr_time.tv_sec - p1->pr_time.tv_sec) == 0)
-#define ORDERKEY_STATE   if ((result = (long) (sorted_state[p2->pr_state] - \
-			       sorted_state[p1->pr_state])) == 0)
+#define ORDERKEY_STATE   if ((result = (long) (sorted_state[(int)p2->pr_state] - \
+			       sorted_state[(int)p1->pr_state])) == 0)
 #define ORDERKEY_PRIO    if ((result = p2->pr_oldpri - p1->pr_oldpri) == 0)
 #define ORDERKEY_RSSIZE  if ((result = p2->pr_rssize - p1->pr_rssize) == 0)
 #define ORDERKEY_MEM     if ((result = (p2->pr_size - p1->pr_size)) == 0)
@@ -1407,12 +1486,13 @@ getptable (struct prpsinfo *baseptr)
     int i;
     struct dirent *direntp;
     struct oldproc *op;
+    hash_pos pos;
+    hash_item_pid *hi;
     static struct timeval lasttime =
 	{0, 0};
     struct timeval thistime;
     double timediff;
     double alpha, beta;
-    struct oldproc *endbase;
 
     gettimeofday (&thistime, NULL);
     /*
@@ -1444,7 +1524,6 @@ getptable (struct prpsinfo *baseptr)
 	beta = 0.5;
     }
 
-    endbase = oldbase + oldprocs;
     currproc = baseptr;
 
     /* before reading /proc files, turn on root privs */
@@ -1475,8 +1554,7 @@ getptable (struct prpsinfo *baseptr)
 	{
 	    (void) close (fd);
 	    continue;
-	}
-       
+	}       
 #else
 	if ((fd = open (direntp->d_name, O_RDONLY)) < 0)
 	    continue;
@@ -1492,7 +1570,9 @@ getptable (struct prpsinfo *baseptr)
 	    /* not a show stopper -- just fill in the needed values */
 	    currproc->pr_fill = 0;
 	    currproc->pr_onpro = 0;
-	} else {
+	}
+	else
+	{
 	    /* copy over the values we need from prstatus */
 	    currproc->pr_fill = (short)prstatus.pr_nlwp;
 	    currproc->pr_onpro = prstatus.pr_processor;
@@ -1504,33 +1584,14 @@ getptable (struct prpsinfo *baseptr)
 	 * to do our own.  See if we've heard of this process before.
 	 * If so, compute % based on CPU since last time.
 	 * NOTE:  Solaris 2.4 and higher do maintain CPU% in prpsinfo.
+	 * ALSO NOTE: the pctcpu recorded in the proc structure does not
+	 * produce the kind of results we really want, so we don't use
+	 * it even though its there.
 	 */
-	op = oldbase + HASH (currproc->pr_pid);
-	while (1)
-	{
-	    if (op->oldpid == -1)	/* not there */
-		break;
-	    if (op->oldpid == currproc->pr_pid)
-	    {			/* found old data */
-#ifndef PROC_HAS_PCTCPU
-		percent_cpu (currproc) =
-		    ((currproc->pr_time.tv_sec * 1.0e9 +
-		      currproc->pr_time.tv_nsec)
-		     - op->oldtime) / timediff;
-#endif
-		weighted_cpu (currproc) =
-		    op->oldpct * beta + percent_cpu (currproc) * alpha;
-
-		break;
-	    }
-	    op++;			/* try next entry in hash table */
-	    if (op == endbase)	/* table wrapped around */
-		op = oldbase;
-	}
-
-	/* Otherwise, it's new, so use all of its CPU time */
-	if (op->oldpid == -1)
+	op = (struct oldproc *)hash_lookup_pid(prochash, currproc->pr_pid);
+	if (op == NULL)
 	{
+	    /* new entry */
 #ifdef PROC_HAS_PCTCPU
 	    weighted_cpu (currproc) =
 		percent_cpu (currproc);
@@ -1550,13 +1611,26 @@ getptable (struct prpsinfo *baseptr)
 	    }
 #endif
 	}
+	else
+	{
+	    /* existing entry */
+#ifndef PROC_HAS_PCTCPU
+	    percent_cpu (currproc) =
+		((currproc->pr_time.tv_sec * 1.0e9 +
+		  currproc->pr_time.tv_nsec)
+		 - op->oldtime) / timediff;
+#endif
+	    weighted_cpu (currproc) =
+		op->oldpct * beta + percent_cpu (currproc) * alpha;
+	}
 
 	numprocs++;
 	currproc = (struct prpsinfo *) ((char *) currproc + PRPSINFOSIZE);
-	(void) close (fd);
+	(void) close(fd);
 
 	/* Atypical place for growth */
-	if (numprocs >= maxprocs) {
+	if (numprocs >= maxprocs)
+	{
 	    reallocproc(2 * numprocs);
 	    currproc = (struct prpsinfo *)
 		((char *)baseptr + PRPSINFOSIZE * numprocs);
@@ -1568,37 +1642,50 @@ getptable (struct prpsinfo *baseptr)
     seteuid(getuid());
 #endif
 
+    dprintf("getptable saw %d procs\n", numprocs);
+
     if (nproc != numprocs)
 	nproc = numprocs;
 
     /*
      * Save current CPU time for next time around
-     * For the moment recreate the hash table each time, as the code
-     * is easier that way.
      */
-    oldprocs = 2 * nproc;
-    endbase = oldbase + oldprocs;
-    for (op = oldbase; op < endbase; op++)
-	op->oldpid = -1;
-    for (i = 0, currproc = baseptr;
-	 i < nproc;
-	 i++, currproc = (struct prpsinfo *) ((char *) currproc + PRPSINFOSIZE))
+    i = 0;
+    currproc = baseptr;
+    while (i++ < nproc)
     {
-	/* find an empty spot */
-	op = oldbase + HASH (currproc->pr_pid);
-	while (1)
+	op = (struct oldproc *)hash_lookup_pid(prochash, currproc->pr_pid);
+	if (op == NULL)
 	{
-	    if (op->oldpid == -1)
-		break;
-	    op++;
-	    if (op == endbase)
-		op = oldbase;
+	    op = (struct oldproc *)malloc(sizeof(struct oldproc));
+	    hash_add_pid(prochash, currproc->pr_pid, (void *)op);
 	}
 	op->oldpid = currproc->pr_pid;
 	op->oldtime = (currproc->pr_time.tv_sec * 1.0e9 +
 		       currproc->pr_time.tv_nsec);
 	op->oldpct = weighted_cpu (currproc);
+	op->seen = 1;
+	currproc = (struct prpsinfo *) ((char *) currproc + PRPSINFOSIZE);
     }
+
+    /* scan the hash table and remove dead entries */
+    hi = hash_first_pid(prochash, &pos);
+    while (hi != NULL)
+    {
+	op = (struct oldproc *)(hi->value);
+	if (op->seen)
+	{
+	    op->seen = 0;
+	}
+	else
+	{
+	    dprintf("removing %d from hash\n", op->oldpid);
+	    hash_remove_pos_pid(&pos);
+	    free(op);
+	}
+	hi = hash_next_pid(&pos);
+    }
+
     lasttime = thistime;
 }
 
@@ -1665,7 +1752,6 @@ setpriority (int dummy, int who, int niceval)
 static void reallocproc(int n)
 {
     int bytes;
-    struct oldproc *op, *endbase;
 
     if (n < maxprocs)
 	return;
@@ -1679,26 +1765,11 @@ static void reallocproc(int n)
     pbase = (struct prpsinfo *) realloc(pbase, bytes);
     pref = (struct prpsinfo **) realloc(pref,
 					maxprocs * sizeof(struct prpsinfo *));
-    oldbase = (struct oldproc *) realloc(oldbase,
-					 2 * maxprocs * sizeof(struct oldproc));
 
     /* Just in case ... */
-    if (pbase == (struct prpsinfo *) NULL || pref == (struct prpsinfo **) NULL
-	|| oldbase == (struct oldproc *) NULL)
+    if (pbase == (struct prpsinfo *) NULL || pref == (struct prpsinfo **) NULL)
     {
 	fprintf (stderr, "%s: can't allocate sufficient memory\n", myname);
 	quit(1);
     }
-
-    /*
-     * We're growing from 0 to some number, only then we need to
-     * init the oldproc stuff
-     */
-    if (!oldprocs) {
-	oldprocs = 2 * maxprocs;
-
-	endbase = oldbase + oldprocs;
-	for (op = oldbase; op < endbase; op++)
-	    op->oldpid = -1;
-    }
 }
