@@ -1,12 +1,38 @@
+/*
+ * Copyright (c) 1984 through 2007, William LeFebvre
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *     * Redistributions of source code must retain the above copyright
+ * notice, this list of conditions and the following disclaimer.
+ * 
+ *     * Redistributions in binary form must reproduce the above
+ * copyright notice, this list of conditions and the following disclaimer
+ * in the documentation and/or other materials provided with the
+ * distribution.
+ * 
+ *     * Neither the name of William LeFebvre nor the names of other
+ * contributors may be used to endorse or promote products derived from
+ * this software without specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  *  Top users/processes display for Unix
  *  Version 3
- *
- *  This program may be freely redistributed,
- *  but this entire comment MUST remain intact.
- *
- *  Copyright (c) 1984, 1989, William LeFebvre, Rice University
- *  Copyright (c) 1989, 1990, 1992, William LeFebvre, Northwestern University
  */
 
 /*
@@ -110,16 +136,21 @@ itoa(int val)
  */
 
 char *
-itoa7(int val)
+itoa_w(int val, int w)
 
 {
-    register char *ptr;
+    char *ptr;
+    char *eptr;
     static char buffer[16];	/* result is built here */
     				/* 16 is sufficient since the largest number
 				   we will ever convert will be 2^32-1,
 				   which is 10 digits. */
 
-    ptr = buffer + sizeof(buffer);
+    if (w > 15)
+    {
+	w = 15;
+    }
+    eptr = ptr = buffer + sizeof(buffer);
     *--ptr = '\0';
     if (val == 0)
     {
@@ -130,13 +161,20 @@ itoa7(int val)
 	*--ptr = (val % 10) + '0';
 	val /= 10;
     }
-    while (ptr > buffer + sizeof(buffer) - 7)
+    while (ptr >= eptr - w)
     {
 	*--ptr = ' ';
     }
     return(ptr);
 }
 
+char *
+itoa7(int val)
+
+{
+    return itoa_w(val, 7);
+}
+
 /*
  *  digits(val) - return number of decimal digits in val.  Only works for
  *	positive numbers.  If val < 0 then digits(val) == 0, but
@@ -173,7 +211,7 @@ printable(char *str)
 
 {
     register char *ptr;
-    register char ch;
+    register int ch;
 
     ptr = str;
     while ((ch = *ptr) != '\0')
@@ -188,12 +226,12 @@ printable(char *str)
 }
 
 /*
- *  strecpy(to, from) - copy string "from" into "to" and return a pointer
+ *  strcpyend(to, from) - copy string "from" into "to" and return a pointer
  *	to the END of the string "to".
  */
 
 char *
-strecpy(char *to, char *from)
+strcpyend(char *to, char *from)
 
 {
     while ((*to++ = *from++) != '\0');
@@ -265,7 +303,7 @@ char *string_list(char **strings)
     int cnt = 0;
     char **pp;
     char *p;
-    char *result;
+    char *result = NULL;
     char *resp = NULL;
 
     pp = strings;
@@ -280,10 +318,10 @@ char *string_list(char **strings)
 	pp = strings;
 	while ((p = *pp++) != NULL)
 	{
-	    resp = strecpy(resp, p);
+	    resp = strcpyend(resp, p);
 	    if (*pp != NULL)
 	    {
-		resp = strecpy(resp, ", ");
+		resp = strcpyend(resp, ", ");
 	    }
 	}
     }
@@ -596,6 +634,80 @@ format_k(long amt)
     return(ret);
 }
 
+/*
+ * Time keeping functions.  
+ */
+
+static struct timeval lasttime = { 0, 0 };
+static unsigned int elapsed_msecs = 0;
+
+void
+time_get(struct timeval *tv)
+
+{
+    /* get the current time */
+#ifdef HAVE_GETTIMEOFDAY
+    gettimeofday(tv, NULL);
+#else
+    tv->tv_sec = (long)time(NULL);
+    tv->tv_usec = 0;
+#endif
+}
+
+void
+time_mark(struct timeval *tv)
+
+{
+    struct timeval thistime;
+    struct timeval timediff;
+
+    /* if the caller didnt provide one then use our own */
+    if (tv == NULL)
+    {
+	tv = &thistime;
+    }
+
+    /* get the current time */
+#ifdef HAVE_GETTIMEOFDAY
+    gettimeofday(tv, NULL);
+#else
+    tv->tv_sec = (long)time(NULL);
+    tv->tv_usec = 0;
+#endif
+
+    /* calculate the difference */
+    timediff.tv_sec = tv->tv_sec - lasttime.tv_sec;
+    timediff.tv_usec = tv->tv_usec - lasttime.tv_usec;
+    if (timediff.tv_usec < 0) {
+	timediff.tv_sec--;
+	timediff.tv_usec += 1000000;
+    }
+
+    /* convert to milliseconds */
+    elapsed_msecs = timediff.tv_sec * 1000 + timediff.tv_usec / 1000;
+    if (elapsed_msecs == 0)
+    {
+	elapsed_msecs = 1;
+    }
+
+    /* save for next time */
+    lasttime = *tv;
+}
+
+unsigned int
+time_elapsed()
+
+{
+    return elapsed_msecs;
+}
+
+unsigned int
+diff_per_second(unsigned int x, unsigned int y)
+
+{
+    return (y > x ? UINT_MAX - y + x + 1 : x - y) * 1000 / elapsed_msecs;
+}
+
 static int debug_on = 0;
 
 #ifdef DEBUG
